import logging
from typing import List, Text, Optional, Dict, Any

from ruamel.yaml.compat import ordereddict
from sqlalchemy import and_
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from rasax.community import config
from rasax.community.database import Role, Permission, Project, User, utils
from rasax.community.services.user_service import UserService

logger = logging.getLogger(__name__)

GUEST_ROLE = "guest"

PERMISSIONS = {
    # This is not a category in the UI, but currently assigned by default.
    # Every user can do this, default permission
    "basic": {
        "view": {"user.get", "chatToken.get", "statistics.get"},
        "modify": {"user.update", "user.password.update"},
    },
    # this cannot be assigned with platform categories currently
    # only admin default role has those permissions
    "advanced": {
        "view": {
            "roles.list",
            "roles.get",
            "roles.users.list",
            "users.list",
            "users.roles.list",
            "allEvaluations.list",
            "domain.get",
            "domainActions.get",
            "domainWarnings.get",
        },
        "modify": {
            "roles.create",
            "roles.update",
            "roles.delete",
            "users.roles.update",
            "users.roles.delete",
            "features.create",
            "users.delete",
            "users.create",
            "projects.create",
            "clientEvaluation.update",
            "clientEvaluation.delete",
            "allEvaluations.create",
            "chatToken.update",
            "domain.update",
        },
    },
    "test_conversation": {
        "view": {
            "clients.get",
            "clientMessages.list",
            "metadata.get",
            # special case falling into view being a create:
            "messages.create",
            "actionPrediction.get",
        }
    },
    "conversations": {
        "view": {
            "metadata.list",
            "logs.list",
            "clientEvaluation.get",
            "conversationActions.list",
            "conversationEntities.list",
            "conversationIntents.list",
            "conversationPolicies.list",
        },
        "modify": {
            "messageFlags.update",
            "messageFlags.delete",
            "metadata.create",
            "actionPrediction.create",
        },
    },
    "nlu_training_data": {
        "view": {
            "examples.list",
            "examples.get",
            "bulkData.get",
            "entities.list",
            "logs.list",
            "logs.get",
            "warnings.get",
            "intents.list",
        },
        "modify": {
            "examples.update",
            "examples.delete",
            "logs.delete",
            "logs.create",
            "examples.create",
            "bulkData.update",
            "messageIntents.update",
            "messageIntents.delete",
            "intents.create",
            "intents.update",
            "intents.delete",
            "userGoals.create",
            "userGoals.delete",
        },
    },
    "stories": {
        "view": {"stories.list", "stories.get", "bulkStories.get"},
        "modify": {
            "stories.create",
            "bulkStories.update",
            "stories.update",
            "stories.delete",
            "clientEvents.create",
            "clientEvents.update",
        },
    },
    "responses": {
        "view": {"responseTemplates.list"},
        "modify": {
            "bulkResponseTemplates.update",
            "responseTemplates.create",
            "responseTemplates.delete",
            "responseTemplates.update",
            "nlgResponse.create",
        },
    },
    "models": {
        "view": {
            "models.list",
            "models.settings.get",
            "models.get",
            "models.modelByTag.get",
            "models.nlu.evaluations.list",
            "models.nlu.evaluations.get",
        },
        "modify": {
            "models.delete",
            "models.tags.update",
            "models.tags.delete",
            "models.jobs.create",
            "models.settings.update",
            "models.nlu.evaluations.delete",
            "models.nlu.evaluations.update",
            "models.create",
        },
    },
    "analytics_dashboard": {"view": {"analytics.get"}},
    "deployment_environment": {
        "view": {"environments.list"},
        "modify": {"environments.update"},
    },
}

DEFAULT_ROLES = {
    "admin": [
        "basic.*",
        "advanced.*",
        "test_conversation.*",
        "conversations.*",
        "nlu_training_data.*",
        "stories.*",
        "responses.*",
        "models.*",
        "analytics_dashboard.*",
        "deployment_environment.*",
    ],
    "annotator": [
        "basic.*",
        "test_conversation.*",
        "nlu_training_data.*",
        "responses.*",
        "stories.*",
    ],
    "tester": ["basic.*", "test_conversation.*"],
}


def guest_permissions() -> List[Text]:
    """Returns list of guest user permissions.

    These are defined by the `basic` and `test_conversation` UI categories.
    """

    basic_permissions = permissions_for_category("basic")
    test_conversation_permissions = permissions_for_category("test_conversation")

    return basic_permissions + test_conversation_permissions


def _category_for_annotation(annotation: Text) -> Optional[Text]:
    """Return base category for `annotation`.

    Example: annotation 'user.get' returns 'basic.view'
    """

    for category, mode_dict in PERMISSIONS.items():
        for mode, permissions in mode_dict.items():
            for permission in permissions:
                if annotation == permission:
                    return ".".join((category, mode))

    return None


def permission_from_route_annotation(annotation: Text) -> Optional[Text]:
    """Return full permission from route annotation.

    Example: annotation 'user.get' returns 'basic.view.user.get'
    """

    category = _category_for_annotation(annotation)
    if category:
        return ".".join((category, annotation))

    raise ValueError("Could not find category for annotation '{}'.".format(annotation))


def permissions_for_category(category: Text) -> List[Text]:
    """Return all permissions associated with base category.

    Example: category 'basic' returns

        ['basic.view.user.get',
        'basic.view.chatToken.get',
        'basic.view.statistics.get',
        'basic.modify.user.update',
        'basic.modify.user.password.update']
    """

    if category not in PERMISSIONS:
        raise ValueError(
            "Could not find category '{}' in PERMISSIONS.".format(category)
        )

    return [
        ".".join((category, k, p)) for k, v in PERMISSIONS[category].items() for p in v
    ]


def normalise_permissions(perms: List[Text]) -> List[Text]:
    """Normalises permission strings to be sanic-jwt-compatible.

    Platform permissions are of the form `a.b.c.action`, but sanic-jwt expects
    `a.b.c:action`.
    """

    out = []
    for p in perms:
        k = p.rfind(".")
        out.append(p[:k] + ":" + p[k + 1 :])

    return out


class RoleService(object):
    def __init__(self, session: Optional[Session] = None):
        self.session = session

    @property
    def default_roles(self) -> Dict[Text, Any]:
        """Dictionary of default Rasa X roles.

        Preserves key order when dumped to yaml.
        """

        return ordereddict(DEFAULT_ROLES)

    @property
    def api_permissions(self) -> List[Text]:
        """Retrieve list of all existing API permissions."""

        permissions_list = []
        for category, actions in PERMISSIONS.items():
            for mode, action_set in actions.items():
                for action in action_set:
                    permissions_list.append(".".join((category, mode, action)))

        return permissions_list

    @property
    def roles(self) -> List[Text]:
        """Retrieve list of all existing API roles."""

        return [r for (r,) in self.session.query(Role.role).distinct().all()]

    def init_roles(
        self, overwrite: bool = False, project_id: Text = config.project_name
    ):
        """Initialize the role system."""

        if not overwrite and self.roles:
            # if there are already roles present,
            # we wont overwrite until forced
            return

        # save roles
        for role in [*DEFAULT_ROLES]:
            self.save_role(role)

        # save permissions
        for role, permissions in self.default_roles.items():
            self.save_permissions_for_role(role, permissions, project_id)

    def save_role(self, role: Text) -> Dict[Text, List[Text]]:
        """Save API role."""

        existing = self.session.query(Role).filter(Role.role == role).first()

        if existing:
            return existing.as_dict()
        else:
            user_role = Role(role=role)
            self.session.add(user_role)
            self.session.commit()

            return user_role.as_dict()

    def delete_role(self, role: Text) -> None:
        """Delete API role."""

        if role == "admin":
            raise ValueError("Admin role cannot be deleted.")

        role = self.session.query(Role).filter(Role.role == role).first()

        self.session.delete(role)

    def _expand_wildcard_permission(self, permission: Text) -> List[Text]:
        """Expand API permission with wildcard syntax (`*`)."""

        permissions_list = []
        if "*" in permission:
            permission = permission.replace("*", "")
            for p in self.api_permissions:
                if p.startswith(permission):
                    permissions_list.append(p)

        elif permission in self.api_permissions:
            permissions_list.append(permission)

        return permissions_list

    def expand_wildcard_permissions(
        self, unexpanded_permissions: List[Text]
    ) -> List[Text]:

        permission_list = set()
        for expanded in map(self._expand_wildcard_permission, unexpanded_permissions):
            permission_list.update(expanded)

        return list(permission_list)

    def get_role_permissions(
        self, role: Text, project_id: Text = config.project_name
    ) -> List[Text]:
        """Retrieve list of API permissions associated with `role`."""

        results = (
            self.session.query(Permission.permission)
            .filter(and_(Permission.role_id == role, Permission.project == project_id))
            .all()
        )
        results = [p for p, in results]

        return self.expand_wildcard_permissions(results)

    def get_user_permissions(
        self,
        username: Text,
        user_service: UserService,
        project_id: Text = config.project_name,
    ) -> List[Text]:
        """Retrieve list of API permissions associated with `username`."""

        user_roles = user_service.fetch_user(username)["roles"]

        user_permissions = set()
        for role in user_roles:
            user_permissions.update(self.get_role_permissions(role, project_id))

        return list(user_permissions)

    def save_permissions_for_role(
        self,
        role: Text,
        permissions_list: List[Text],
        project_id: Text = config.project_name,
    ) -> Dict[Text, Any]:
        """Save API permissions for `role`."""

        project_id, = (
            self.session.query(Project.project_id)
            .filter(Project.project_id == project_id)
            .first()
        )

        if project_id:
            permission_objects = [
                Permission(role_id=role, permission=p, project=project_id)
                for p in permissions_list
            ]

            try:
                self.session.bulk_save_objects(permission_objects)
                self.session.commit()
            except IntegrityError as e:
                logger.debug(
                    "Permissions '{}' could not be saved due to '{}'."
                    "".format(permission_objects, e)
                )
                self.session.rollback()

        return {
            "role": role,
            "permissions": self.get_role_permissions(role, project_id),
        }

    def delete_permissions_for_role(
        self, role: Text, project_id: Text = config.project_name
    ) -> None:
        """Delete API permissions for `role`."""

        permissions = (
            self.session.query(Permission)
            .filter(and_(Permission.role_id == role, Permission.project == project_id))
            .all()
        )
        utils.bulk_delete(permissions, self.session)
        self.session.commit()

    def backend_to_frontend_format_roles(
        self, backend_roles: List[Text]
    ) -> List[Dict[Text, Any]]:
        """Translate roles from backend format into the one expected
       by the frontend.

       input format: List[category.mode.specifics] (for example: basic.view.roles.get)
       output format: List[{"role": "role_name",
                            "grants": {"categoryA":['read:any'],
                                    "categoryB": ...}}]
       """

        crud_any = ["create:any", "update:any", "delete:any", "read:any"]

        # first two categories are not used (not frontend categories)
        ui_categories = [*PERMISSIONS][2:]
        roles = []
        for role in backend_roles:
            granted = {}
            permissions_list = self.get_role_permissions(role)

            # check for each permission category if granted
            # if "modify" permission is given, "view" is added automatically
            for permission in permissions_list:
                for item in ui_categories:
                    if permission.startswith(item + ".view") and item not in granted:
                        granted[item] = ["read:any"]
                    elif permission.startswith(item):
                        granted[item] = crud_any

            roles.append({"role": role, "grants": granted})

        return roles

    def frontend_to_backend_format_permissions(
        self, frontend_permissions: Dict[Text, List]
    ) -> List[Text]:
        """Translate permissions from frontend format into the one expected
        by the backend.

        input format: ordereddict([('categoryA', ['read:any']), ...])
        output format: List[category.mode.specifics] (for example: basic.view.roles.get)
        """

        granted = []
        for category, permissions in frontend_permissions.items():
            if "create:any" in permissions:  # "modify" permission
                granted.append(category + ".*")
            else:
                granted.append(category + ".view.*")

            granted = self.expand_wildcard_permissions(granted)

        return granted

    def fetch_role_users(self, role: Text) -> List[Dict[Text, Any]]:
        """Retrieve list of all existing usernames for role."""

        users = self.session.query(User).filter(User.roles.any(Role.role == role)).all()

        return [user.as_dict() for user in users]

    def count_role_users(self, role: Text) -> int:
        """Retrieve number of all existing users for role."""

        return (
            self.session.query(User).filter(User.roles.any(Role.role == role)).count()
        )

    def update_role(
        self,
        role: Text,
        name: Text,
        frontend_permissions: [Dict[Text, List], None],
        user_service: UserService,
    ) -> None:
        """Update role with new name and/ or new permissions."""

        if name != role:
            self._update_role_name(name, role, user_service)

        elif role not in self.roles:
            self.save_role(role)

        elif frontend_permissions:  # delete existing permissions if a new set is passed
            self.delete_permissions_for_role(role)

        if frontend_permissions:
            permissions = self.frontend_to_backend_format_permissions(
                frontend_permissions
            )
            permissions.extend(self.expand_wildcard_permissions(["basic.*"]))
            self.save_permissions_for_role(name, permissions)

    def _update_role_name(
        self, new_name: Text, old_name: Text, user_service: UserService
    ) -> None:
        """Update the name of a role and move its users."""

        self.save_role(new_name)
        # move users with old role name to new role name
        users = self.fetch_role_users(old_name)
        for user in users:
            user_service.replace_user_roles(user["username"], [new_name])
        if old_name in self.roles:
            self.delete_role(old_name)

    def create_role(self, role: Text, frontend_permissions: Dict[Text, List]) -> None:
        """Create a new role with its permissions."""

        backend_permissions = self.frontend_to_backend_format_permissions(
            frontend_permissions
        )

        # all roles receive basic permissions
        backend_permissions.extend(self.expand_wildcard_permissions(["basic.*"]))

        self.save_role(role)
        self.save_permissions_for_role(role, backend_permissions)
